@@ -2,7 +2,7 @@ from .softmax import Softmax
 from .matmul import MatMul, Linear
 from .conv import _sparse_conv2d, Conv2d
 from .attention import MultiheadAttention
-from .simplifiedsparseattention import SimplifiedSparseAttention
+from .deepspeedsparseselfattention import DeepSpeedSparseSelfAttention, SparsityConfig
 from .batchnorm import BatchNorm2d
 from .permute import _permute, Permute
 from .relu import ReLU
