@@ -1,6 +1,7 @@
 import triton
 import torch
 import math
+import torch_blocksparse_cpp_utils
 
 src = '''
   __global__ void NAME (TYPE* A __readonly  __noalias __aligned(16),
@@ -211,7 +212,7 @@ class _sparse_matmul(torch.autograd.Function):
   sdd_cache = dict()
   dsd_cache = dict()
   dds_cache = dict()
-  locks = None
+  locks = dict()
 
   # Given an array sizes representing reduction size for each
   # column of a block-mode matrix multiplication,
@@ -266,141 +267,19 @@ class _sparse_matmul(torch.autograd.Function):
     return segments, column, lockid, maxid, offsets
   
   @staticmethod
-  def get_locks(size):
-    if _sparse_matmul.locks is None or \
-        size > _sparse_matmul.locks.size(0):
-      _sparse_matmul.locks = torch.zeros(size, dtype=torch.int32).cuda()
-    return _sparse_matmul.locks
+  def get_locks(size, dev):
+    if dev not in _sparse_matmul.locks or \
+        size > _sparse_matmul.locks[dev].size(0):
+      _sparse_matmul.locks[dev] = torch.zeros(size, dtype=torch.int32, device=dev)
+    return _sparse_matmul.locks[dev]
 
   ##########################
   # SPARSE = DENSE x DENSE #
   ##########################
-  _sdd_segment_src = '''
-
-#ifdef _OPENMP
-#include <omp.h>
-#endif
-
-typedef std::vector<std::tuple<int, torch::Tensor>> ret_t;
-
-void segment_blocks(torch::Tensor layout, torch::Tensor idx, torch::Tensor scratch, int max_width, ret_t& ret){
-  size_t H = layout.size(0);
-  size_t M = layout.size(1);
-  size_t N = layout.size(2);
-  torch::Tensor tmp = torch::zeros_like(layout);
-
-  
-  auto _tmp     = tmp.accessor    <int, 3>();
-  auto _layout  = layout.accessor <int, 3>();
-  auto _idx     = idx.accessor    <int, 3>();
-  auto _scratch = scratch.accessor<int, 3>();
-  std::vector<int> current(H, 0);
-
-
-  #ifdef _OPENMP
-  #pragma omp parallel for
-  #endif
-  for(size_t h = 0; h < H; h++){
-
-    // surrounding indices
-    std::vector<int>              ii_left(max_width, -1);
-    std::vector<std::vector<int>> ii_top(max_width, std::vector<int>(N, -1));
-
-    for(size_t m = 0; m < M; m++){
-      for(size_t n = 0; n < N; n++){
-        int v     = _layout[h][m][n];
-        if(v == 0)
-          continue;
-        int n_left= ii_left[max_width-1];
-        int m_top = ii_top [max_width-1][n];
-        int top      = (m_top >= 0)               ? _tmp[h][m_top][n]      : 0;
-        int left     = (n_left >= 0)              ? _tmp[h][m][n_left]     : 0;
-        int topleft  = (m_top >=0 && n_left >= 0) ? _tmp[h][m_top][n_left] : 0;
-        int width    = std::min(left, std::min(top, topleft)) + 1;
-
-        // reset width if blocks cannot be 
-        // packed together (i.e., there's a 1 "in the middle")
-        for(int nn = n_left + 1; nn < n; nn++)
-          if(ii_top[max_width-1][nn] > ii_top[max_width-1][n])
-            width = 1;
-        _tmp[h][m][n] = width;
-
-        // update n_left ring buffer
-        for(int k = 0; k < max_width-1; k++)
-          ii_left[k] = ii_left[k+1];
-        ii_left[max_width-1] = n;
-
-        // update ii_top ring buffer
-        for(int k = 0; k < max_width-1; k++)
-          ii_top[k][n] = ii_top[k+1][n];
-        ii_top[max_width-1][n] = m;
-
-        // block is too small -- skip
-        if(width != max_width)
-          continue;
-
-        // retained blocks are set to zeros
-        for(size_t km = 0; km < max_width; km++)
-        for(size_t kn = 0; kn < max_width; kn++)
-        {
-          int mm = ii_top[km][n];
-          int nn = ii_left[kn];
-          if(mm < 0 || nn < 0)
-            continue;
-          _layout[h][mm][nn] = 0;
-          _tmp[h][mm][nn] = 0;
-          _scratch[h][current[h]][0] = (int)h;
-          _scratch[h][current[h]][1] = (int)mm;
-          _scratch[h][current[h]][2] = (int)nn;
-          _scratch[h][current[h]][3] = _idx[h][mm][nn];
-          current[h]++;
-        }
-      }
-    }
-  }
-  std::vector<torch::Tensor> to_cat;
-  for(size_t h = 0; h < H; h++)
-    if(current[h] > 0)
-      to_cat.push_back(scratch[h].slice(0, 0, current[h]));
-  if(!to_cat.empty())
-    ret.push_back({max_width, torch::cat(to_cat)});
-}
-
-ret_t sdd_segment(torch::Tensor layout, int start_width) {
-  ret_t ret;
-
-  // block index
-  torch::Tensor idx = torch::zeros_like(layout);
-  int current = 0;
-  size_t H = layout.size(0);
-  size_t M = layout.size(1);
-  size_t N = layout.size(2);
-  auto _layout  = layout.accessor <int, 3>();
-  auto _idx = idx.accessor<int, 3>();
-  for(size_t h = 0; h < H; h++)
-  for(size_t m = 0; m < M; m++)
-  for(size_t n = 0; n < N; n++){
-    if(_layout[h][m][n] == 0)
-      continue;
-    _idx[h][m][n] = current++;
-  }
-
-  // scratch memory
-  torch::Tensor scratch = torch::empty({H, layout.sum().item<int>(), 4}, layout.dtype());
-
-  for(int max_width = start_width; max_width > 0; max_width /= 2)
-    segment_blocks(layout, idx, scratch, max_width, ret);
-  return ret;
-}
-'''
-
-  sdd_segment = torch.utils.cpp_extension.load_inline(name='sdd_segment', 
-                                                      cpp_sources=[_sdd_segment_src], 
-                                                      functions=['sdd_segment'],
-                                                      extra_cflags=['-O2', '-fopenmp']).sdd_segment
+  sdd_segment = torch_blocksparse_cpp_utils.sdd_segment
 
   @staticmethod
-  def make_sdd_lut(layout, block, dtype):
+  def make_sdd_lut(layout, block, dtype, device):
     start_width = 64 // block
     segmented = _sparse_matmul.sdd_segment(layout.type(torch.int32), start_width)
     luts, widths, packs = [], [], []
@@ -411,7 +290,7 @@ ret_t sdd_segment(torch::Tensor layout, int start_width) {
       j = nnz[:, 2]
       b = nnz[:, 3]
       lut = torch.stack((h, i, j, b), dim=1).view(-1).contiguous()
-      luts.append(lut.type(torch.int32).cuda()) 
+      luts.append(lut.type(torch.int32).to(device)) 
       widths.append(width)
       packs.append(size)
     # create locks
@@ -433,15 +312,24 @@ ret_t sdd_segment(torch::Tensor layout, int start_width) {
     BS2 = b.size(3 if trans_b else 2)
     BS3 = b.size(2 if trans_b else 3)
     dtype = a.dtype
+    is_16_multiple = AS3 % 16 == 0
+    is_32_multiple = AS3 % 32 == 0
+    is_64_multiple = AS3 % 64 == 0
+    if not is_16_multiple:
+      raise ValueError('Reduction size for SDD must be a multiple of 16')
     # create kernel
     total_width = sum([width*pack*pack for width,pack in zip(widths, packs)])
     c = torch.empty((AS0, total_width, block, block), dtype=dtype, device=a.device)
     for lut, width, pack in zip(luts, widths, packs):
       num_lock = 1
-      key = (block, a.dtype, b.dtype, trans_a, trans_b, trans_c, pack)
+      key = (block, a.dtype, b.dtype, trans_a, trans_b, trans_c, pack, is_32_multiple, is_64_multiple)
       if key not in _sparse_matmul.sdd_cache:
-        TK = {torch.float32: [8, 16],
-              torch.float16: [16, 32, 64]}[dtype]
+        F32TK  = [8, 16]
+        F16TK  = [16]
+        F16TK += [32] if is_32_multiple else []
+        F16TK += [64] if is_64_multiple else []
+        TK = {torch.float32: F32TK,
+              torch.float16: F16TK}[dtype]
         defines =  {'TM': block*pack, 'TN': block*pack, 'TMN': block*block*pack*pack, 'BLOCK': block, 
                   'TK': TK, 'TYPE': dtype,
                   'STRIDE_AM': '1'    if trans_a else 'lda', 
@@ -455,7 +343,7 @@ ret_t sdd_segment(torch::Tensor layout, int start_width) {
 
       kernel = _sparse_matmul.sdd_cache[key]
       # create output
-      locks = _sparse_matmul.get_locks(2*width*AS0*num_lock)
+      locks = _sparse_matmul.get_locks(2*width*AS0*num_lock, a.device)
       # maximum grid size is 65535
       # so operation might be decomposed into multiple
       # kernel calls
@@ -481,7 +369,7 @@ ret_t sdd_segment(torch::Tensor layout, int start_width) {
   # Given a binary layout of 0s and 1s,
   # Construct look-up table for efficient execution on GPUs
   @staticmethod
-  def make_dxx_lut(layout, block, step, trans, transform = lambda idx: idx):
+  def make_dxx_lut(layout, block, step, trans, device, transform = lambda idx: idx):
     # load-balancing
     _empty = torch.tensor([], dtype=torch.int64, device=layout.device)
     segments = _empty.clone()
@@ -560,9 +448,10 @@ ret_t sdd_segment(torch::Tensor layout, int start_width) {
     offsets += 6*width
     header = torch.stack((offsets, segments, column, depth, lockid, maxid), dim=1).view(-1).contiguous()
     incs = torch.stack((xincs, wincs), dim=1).view(-1).contiguous()
+    incs = torch.cat((incs, torch.zeros(2, device=incs.device, dtype=incs.dtype)))
     # create lut
     lut = torch.cat((header, incs))
-    lut = lut.type(torch.int32).cuda()
+    lut = lut.type(torch.int32).to(device)
     # create locks
     num_locks = max(1, lockid.max())
     return lut, num_locks, width, None
@@ -603,7 +492,7 @@ ret_t sdd_segment(torch::Tensor layout, int start_width) {
     CS1 = AS1
     CS2 = BS2 if trans_c else AS2
     CS3 = AS2 if trans_c else BS2
-    locks = _sparse_matmul.get_locks(2*AS0*AS2//32*num_locks)
+    locks = _sparse_matmul.get_locks(2*AS0*AS2//32*num_locks, a.device)
     c = torch.empty((CS0, CS1, CS2, CS3), dtype=dtype, device=a.device)
     time[0] = kernel(a, b, c, 
                      a.stride(2), block, c.stride(2), 
@@ -650,7 +539,7 @@ ret_t sdd_segment(torch::Tensor layout, int start_width) {
     CS1 = BS1
     CS2 = BS3 if trans_c else AS1
     CS3 = AS1 if trans_c else BS3
-    locks = _sparse_matmul.get_locks(2*BS0*BS3//32*num_locks)
+    locks = _sparse_matmul.get_locks(2*BS0*BS3//32*num_locks, a.device)
     c = torch.empty((CS0, CS1, CS2, CS3), dtype=dtype, device=a.device)
     time[0] = kernel(a, b, c, 
                      block, b.stride(2), c.stride(2), 
@@ -718,33 +607,33 @@ ret_t sdd_segment(torch::Tensor layout, int start_width) {
 class MatMul:
   
   
-  def make_lut(self, dtype):
-    key = (dtype, )
+  def make_lut(self, dtype, device):
+    key = (dtype, device)
     if key in self.lut_cache:
       return self.lut_cache[key]
     # C look-up table
     layout, block = self.layout, self.block
     step = 8 if dtype == torch.float32 else 16
     if self.mode == 'sdd':
-      c_lut, c_num_locks, c_width, c_packs = _sparse_matmul.make_sdd_lut(layout, block, dtype)
+      c_lut, c_num_locks, c_width, c_packs = _sparse_matmul.make_sdd_lut(layout, block, dtype, device)
     elif self.mode == 'dsd':
-      c_lut, c_num_locks, c_width, c_packs = _sparse_matmul.make_dxx_lut(layout, block, step, not self.trans_a)
+      c_lut, c_num_locks, c_width, c_packs = _sparse_matmul.make_dxx_lut(layout, block, step, not self.trans_a, device)
     elif self.mode == 'dds':
-      c_lut, c_num_locks, c_width, c_packs = _sparse_matmul.make_dxx_lut(layout, block, step, self.trans_b)
+      c_lut, c_num_locks, c_width, c_packs = _sparse_matmul.make_dxx_lut(layout, block, step, self.trans_b, device)
     # DA look-up table
     if self.mode == 'sdd':
-      da_lut, da_num_locks, da_width, da_packs = _sparse_matmul.make_dxx_lut(layout, block, step, True)
+      da_lut, da_num_locks, da_width, da_packs = _sparse_matmul.make_dxx_lut(layout, block, step, True, device)
     elif self.mode == 'dsd':
-      da_lut, da_num_locks, da_width, da_packs = _sparse_matmul.make_sdd_lut(layout, block, dtype)
+      da_lut, da_num_locks, da_width, da_packs = _sparse_matmul.make_sdd_lut(layout, block, dtype, device)
     elif self.mode == 'dds':
-      da_lut, da_num_locks, da_width, da_packs = _sparse_matmul.make_dxx_lut(layout, block, step, not self.trans_b)
+      da_lut, da_num_locks, da_width, da_packs = _sparse_matmul.make_dxx_lut(layout, block, step, not self.trans_b, device)
     # DB look-up table
     if self.mode == 'sdd':
-      db_lut, db_num_locks, db_width, db_packs = _sparse_matmul.make_dxx_lut(layout, block, step, False)
+      db_lut, db_num_locks, db_width, db_packs = _sparse_matmul.make_dxx_lut(layout, block, step, False, device)
     elif self.mode == 'dsd':
-      db_lut, db_num_locks, db_width, db_packs = _sparse_matmul.make_dxx_lut(layout, block, step, self.trans_a)
+      db_lut, db_num_locks, db_width, db_packs = _sparse_matmul.make_dxx_lut(layout, block, step, self.trans_a, device)
     elif self.mode == 'dds':
-      db_lut, db_num_locks, db_width, db_packs = _sparse_matmul.make_sdd_lut(layout, block, dtype)
+      db_lut, db_num_locks, db_width, db_packs = _sparse_matmul.make_sdd_lut(layout, block, dtype, device)
     self.lut_cache[key] = (c_lut, c_num_locks, c_width, c_packs,\
                            da_lut, da_num_locks, da_width, da_packs,\
                            db_lut, db_num_locks, db_width, db_packs)
@@ -780,7 +669,7 @@ class MatMul:
   def __call__(self, a, b):
     c_lut, c_num_locks, c_width, c_packs,\
     da_lut, da_num_locks, da_width, da_packs,\
-    db_lut, db_num_locks, db_width, db_packs = self.make_lut(a.dtype)
+    db_lut, db_num_locks, db_width, db_packs = self.make_lut(a.dtype, a.device)
     # timings
     time_c  = [None]
     time_da = [None]
@@ -815,4 +704,4 @@ class Linear(torch.nn.Module):
     torch.nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
     
   def forward(self, input):
-    return self.matmul(input, self.weight)
\ No newline at end of file
+    return self.matmul(input, self.weight)
